@@ -34,6 +34,22 @@ typedef struct APXSERVICE {
 
 } APXSERVICE, *LPAPXSERVICE;
 
+/** Maps dwCurrentState to the constant name described on https://docs.microsoft.com/en-us/windows/win32/api/winsvc/ns-winsvc-service_status */
+static const char* gSzCurrentState[] = {
+    "",
+    "SERVICE_STOPPED",
+    "SERVICE_START_PENDING",
+    "SERVICE_STOP_PENDING",
+    "SERVICE_RUNNING",
+    "SERVICE_CONTINUE_PENDING",
+    "SERVICE_PAUSE_PENDING",
+    "SERVICE_PAUSED"
+};
+
+const char* apxServiceGetStateName(DWORD dwCurrentState) {
+    return gSzCurrentState[dwCurrentState < 0 ? 0 : dwCurrentState > _countof(gSzCurrentState) ? 0 : dwCurrentState];
+}
+
 static WCHAR __invalidPathChars[] = L"<>:\"/\\:|?*";
 static BOOL __apxIsValidServiceName(LPCWSTR szServiceName)
 {
@@ -73,6 +89,8 @@ static BOOL __apxServiceCallback(APXHANDLE hObject, UINT uMsg,
         break;
     }
     return TRUE;
+	UNREFERENCED_PARAMETER(wParam);
+	UNREFERENCED_PARAMETER(lParam);
 }
 
 APXHANDLE
@@ -138,13 +156,13 @@ apxServiceOpen(APXHANDLE hService, LPCWSTR szServiceName, DWORD dwOptions)
     if (!apxGetServiceDescriptionW(szServiceName,
                                    lpService->stServiceEntry.szServiceDescription,
                                    SIZ_DESLEN)) {
-        apxLogWrite(APXLOG_MARK_WARN "Failed to obtain service description");
+        apxLogWrite(APXLOG_MARK_WARN "Failed to obtain service description for '%s'", szServiceName);
         lpService->stServiceEntry.szServiceDescription[0] = L'\0';
     }
     if (!apxGetServiceUserW(szServiceName,
                             lpService->stServiceEntry.szObjectName,
                             SIZ_RESLEN)) {
-        apxLogWrite(APXLOG_MARK_WARN "Failed to obtain service user name");
+        apxLogWrite(APXLOG_MARK_WARN "Failed to obtain service user name for '%s'", szServiceName);
         lpService->stServiceEntry.szObjectName[0] = L'\0';
     }
     if (!QueryServiceConfigW(lpService->hService, NULL, 0, &dwNeeded)) {
@@ -236,10 +254,8 @@ apxServiceSetNames(APXHANDLE hService,
     /* Check if the ServiceOpen has been called */
     if (IS_INVALID_HANDLE(lpService->hService))
         return FALSE;
-    if (szUsername || szPassword)
-        dwServiceType = SERVICE_WIN32_OWN_PROCESS;
     if (!ChangeServiceConfigW(lpService->hService,
-                              dwServiceType,
+                              SERVICE_NO_CHANGE,
                               SERVICE_NO_CHANGE,
                               SERVICE_NO_CHANGE,
                               szImagePath,
@@ -264,37 +280,55 @@ apxServiceSetNames(APXHANDLE hService,
 
 BOOL
 apxServiceSetOptions(APXHANDLE hService,
-                     DWORD dwServiceType,
-                     DWORD dwStartType,
-					 BOOL bDelayedStart,
-                     DWORD dwErrorControl)
+    LPCWSTR lpDependencies,
+    DWORD dwServiceType,
+    DWORD dwStartType,
+    BOOL bDelayedStart,
+    DWORD dwErrorControl)
 {
     LPAPXSERVICE lpService;
     SERVICE_DELAYED_AUTO_START_INFO sDelayedInfo;
 
-    if (hService->dwType != APXHANDLE_TYPE_SERVICE)
+    if (hService->dwType != APXHANDLE_TYPE_SERVICE) {
+        apxLogWrite(APXLOG_MARK_WARN "Can't set options for service.");
         return FALSE;
+    }
 
     lpService = APXHANDLE_DATA(hService);
     /* Manager mode cannot handle services */
-    if (lpService->bManagerMode)
+    if (lpService->bManagerMode) {
+        apxLogWrite(APXLOG_MARK_WARN "Can't set options for service: Manager mode cannot handle services");
         return FALSE;
-    /* Check if the ServixeOpen has been called */
-    if (IS_INVALID_HANDLE(lpService->hService))
+    }
+    /* Check if the ServiceOpen has been called */
+    if (IS_INVALID_HANDLE(lpService->hService)) {
+        apxLogWrite(APXLOG_MARK_WARN "Can't set options for service: Service is not open.");
         return FALSE;
+    }
+
+    /* Add the mandatory dependencies */
+    if (lpDependencies) {
+        lpDependencies = apxMultiSzCombine(NULL, lpDependencies,
+                                           L"Tcpip\0Afd\0", NULL);
+    } else {
+        lpDependencies = L"Tcpip\0Afd\0";
+    }
 
     if (!ChangeServiceConfig(lpService->hService, dwServiceType,
                                    dwStartType, dwErrorControl,
-                                   NULL, NULL, NULL, NULL, NULL,
-                                   NULL, NULL)) {
+                                   NULL, NULL, NULL,
+                                   lpDependencies,
+                                   NULL, NULL, NULL)) {
+        apxLogWrite(APXLOG_MARK_WARN "Can't set options for service: Failed to change the configuration parameters.");
     	return FALSE;
     }
 
     if (dwStartType == SERVICE_AUTO_START) {
     	sDelayedInfo.fDelayedAutostart = bDelayedStart;
-    	return ChangeServiceConfig2A(lpService->hService,
-                                     SERVICE_CONFIG_DELAYED_AUTO_START_INFO,
-                                     &sDelayedInfo);
+        if (!ChangeServiceConfig2A(lpService->hService, SERVICE_CONFIG_DELAYED_AUTO_START_INFO, &sDelayedInfo)) {
+            apxLogWrite(APXLOG_MARK_WARN "Can't set options for service: Failed to change the optional configuration parameters.");
+            return FALSE;
+        }
     }
 
     return TRUE;
@@ -407,7 +441,7 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
 {
     LPAPXSERVICE   lpService;
     SERVICE_STATUS stStatus;
-    DWORD          dwPending = 0;
+    DWORD          dwPendingState = 0;
     DWORD          dwState = 0;
     DWORD          dwTick  = 0;
     DWORD          dwWait, dwCheck, dwStart, sleepMillis;
@@ -429,16 +463,16 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
     }
     switch (dwControl) {
         case SERVICE_CONTROL_CONTINUE:
-            dwPending = SERVICE_START_PENDING;
-            dwState   = SERVICE_RUNNING;
+            dwPendingState = SERVICE_START_PENDING;
+            dwState        = SERVICE_RUNNING;
             break;
         case SERVICE_CONTROL_STOP:
-            dwPending = SERVICE_STOP_PENDING;
-            dwState   = SERVICE_STOPPED;
+            dwPendingState = SERVICE_STOP_PENDING;
+            dwState        = SERVICE_STOPPED;
             break;
         case SERVICE_CONTROL_PAUSE:
-            dwPending = SERVICE_PAUSE_PENDING;
-            dwState   = SERVICE_PAUSED;
+            dwPendingState = SERVICE_PAUSE_PENDING;
+            dwState        = SERVICE_PAUSED;
             break;
         default:
             break;
@@ -452,28 +486,30 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
         switch (dwControl & 0xE0) {
             case 0x80:
             case 0x90:
-                dwPending = SERVICE_START_PENDING;
-                dwState   = SERVICE_RUNNING;
+                dwPendingState = SERVICE_START_PENDING;
+                dwState        = SERVICE_RUNNING;
                 break;
             case 0xA0:
             case 0xB0:
-                dwPending = SERVICE_STOP_PENDING;
-                dwState   = SERVICE_STOPPED;
+                dwPendingState = SERVICE_STOP_PENDING;
+                dwState        = SERVICE_STOPPED;
                 break;
             case 0xC0:
             case 0xD0:
-                dwPending = SERVICE_PAUSE_PENDING;
-                dwState   = SERVICE_PAUSED;
+                dwPendingState = SERVICE_PAUSE_PENDING;
+                dwState        = SERVICE_PAUSED;
                 break;
             default:
                 break;
         }
     }
-    if (!dwPending && !dwState) {
+    if (!dwPendingState && !dwState) {
         apxLogWrite(APXLOG_MARK_ERROR
-            "apxServiceControl():  !dwPending(%d) && !dwState(%d); returning FALSE",
-            dwPending,
-            dwState);
+            "apxServiceControl():  !dwPendingState(%d = %s) && !dwState(%d = %s); returning FALSE",
+            dwPendingState,
+            apxServiceGetStateName(dwPendingState),
+            dwState,
+            apxServiceGetStateName(dwState));
         return FALSE;
     }
     /* Now lets control */
@@ -503,7 +539,7 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
     if (bStatus) {
         Sleep(100); /* Initial Sleep period */
         while (QueryServiceStatus(lpService->hService, &stStatus)) {
-            if (stStatus.dwCurrentState == dwPending) {
+            if (stStatus.dwCurrentState == dwPendingState) {
                 /* Do not wait longer than the wait hint. A good interval is
                  * one tenth the wait hint, but no less than 1 second and no
                  * more than 10 seconds.
@@ -539,7 +575,7 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
                 break;
         }
     }
-    /* signal that we are done with controling the service */
+    /* signal that we are done with controlling the service */
     if (fnControlCallback)
         (*fnControlCallback)(lpCbData, uMsg, (WPARAM)3, (LPARAM)0);
     /* Check if we are in the desired state */
@@ -558,10 +594,12 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
             return TRUE;
         } else {
             apxLogWrite(APXLOG_MARK_ERROR
-                "apxServiceControl(): dwState(%d) != dwCurrentState(%d); "
-                "dwWin32ExitCode = %d, dwWaitHint = %d, dwServiceSpecificExitCode = %d",
+                "apxServiceControl(): dwState(%d = %s) != dwCurrentState(%d = %s); "
+                "dwWin32ExitCode = %d, dwWaitHint = %d millseconds, dwServiceSpecificExitCode = %d",
                 dwState,
+                apxServiceGetStateName(dwState),
                 stStatus.dwCurrentState,
+                apxServiceGetStateName(stStatus.dwCurrentState),
                 stStatus.dwWin32ExitCode,
                 stStatus.dwWaitHint,
                 stStatus.dwServiceSpecificExitCode);
@@ -575,6 +613,58 @@ apxServiceControl(APXHANDLE hService, DWORD dwControl, UINT uMsg,
     return FALSE;
 }
 
+/* Wait one second and check that the service has stopped, returns TRUE if stopped FALSE otherwise */
+BOOL
+apxServiceCheckStop(APXHANDLE hService)
+{
+    LPAPXSERVICE   lpService;
+    SERVICE_STATUS stStatus;
+    DWORD          dwState = SERVICE_STOPPED;
+    DWORD          sleepMillis;
+
+    if (hService->dwType != APXHANDLE_TYPE_SERVICE)
+        return FALSE;
+
+    lpService = APXHANDLE_DATA(hService);
+    /* Manager mode cannot handle services */
+    if (lpService->bManagerMode) {
+        apxLogWrite(APXLOG_MARK_ERROR "apxServiceCheckStop(): Manager mode cannot handle services, returning FALSE");
+        return FALSE;
+    }
+    /* Check if the ServiceOpen has been called */
+    if (IS_INVALID_HANDLE(lpService->hService)) {
+        apxLogWrite(APXLOG_MARK_ERROR "apxServiceCheckStop(): Service is not open, returning FALSE");
+        return FALSE;
+    }
+
+    /* Check if we are in the stopped state */
+    sleepMillis = 1000;
+    apxLogWrite(APXLOG_MARK_DEBUG "apxServiceCheckStop(): Sleeping %d milliseconds", sleepMillis);
+    Sleep(sleepMillis);
+
+    if (QueryServiceStatus(lpService->hService, &stStatus)) {
+        apxLogWrite(APXLOG_MARK_DEBUG "apxServiceCheckStop(): QueryServiceStatus OK");
+        if (stStatus.dwCurrentState == dwState) {
+            return TRUE;
+        } else {
+            apxLogWrite(APXLOG_MARK_DEBUG
+                "apxServiceCheckStop(): dwState(%d) != dwCurrentState(%d); "
+                "dwWin32ExitCode = %d, dwWaitHint = %d milliseconds, dwServiceSpecificExitCode = %d",
+                dwState,
+                stStatus.dwCurrentState,
+                stStatus.dwWin32ExitCode,
+                stStatus.dwWaitHint,
+                stStatus.dwServiceSpecificExitCode);
+        }
+
+    } else {
+        apxLogWrite(APXLOG_MARK_ERROR "apxServiceCheckStop(): QueryServiceStatus failure");
+    }
+
+    apxLogWrite(APXLOG_MARK_DEBUG "apxServiceCheckStop(): returning FALSE");
+    return FALSE;
+}
+
 BOOL
 apxServiceInstall(APXHANDLE hService, LPCWSTR szServiceName,
                   LPCWSTR szDisplayName, LPCWSTR szImagePath,
@@ -582,6 +672,7 @@ apxServiceInstall(APXHANDLE hService, LPCWSTR szServiceName,
                   DWORD dwStartType)
 {
     LPAPXSERVICE   lpService;
+    LPCWSTR szServiceUser = L"NT Authority\\LocalService";
 
     if (hService->dwType != APXHANDLE_TYPE_SERVICE)
         return FALSE;
@@ -608,6 +699,13 @@ apxServiceInstall(APXHANDLE hService, LPCWSTR szServiceName,
                                            L"Tcpip\0Afd\0", NULL);
     else
         lpDependencies = L"Tcpip\0Afd\0";
+
+    if ((dwServiceType & SERVICE_INTERACTIVE_PROCESS) == SERVICE_INTERACTIVE_PROCESS) {
+        // Caller is responsible for checking the user is set appropriately
+        // for an interactive service (will use LocalSystem)
+        szServiceUser = NULL;
+    }
+
     lpService->hService = CreateServiceW(lpService->hManager,
                                          szServiceName,
                                          szDisplayName,
@@ -619,11 +717,12 @@ apxServiceInstall(APXHANDLE hService, LPCWSTR szServiceName,
                                          NULL,
                                          NULL,
                                          lpDependencies,
-                                         L"NT Authority\\LocalService",
+                                         szServiceUser,
                                          NULL);
 
     if (IS_INVALID_HANDLE(lpService->hService)) {
         apxLogWrite(APXLOG_MARK_SYSERR);
+        SetLastError(0);
         return FALSE;
     }
     else {
